/****************************************************************
 * fboctree.c:
 ****************************************************************/

/******
  Copyright (C) 1993-1995 by Klaus Ehrenfried. 

  Permission to use, copy, modify, and distribute this software
  is hereby granted, provided that the above copyright notice appears 
  in all copies and that the software is available to all free of charge. 
  The author disclaims all warranties with regard to this software, 
  including all implied warranties of merchant-ability and fitness. 
  The code is simply distributed as it is.
  *******/

/*******
  This program is based on the Octree algorithm described 
  by Michael Gervautz and Werner Purgathofer (Technical University 
  Vienna, Austria) in the article "A Simple Method for Color Quantization:
  Octree Quantization" published in Graphic Gems edited by Andrew Glassner
  pp. 287 ff. By default a modified version of the octree algorithm 
  is used, but also the original version is implemented.
  *******/

#include <stdio.h>
#include <math.h>
#include <string.h>

#include "fbm.h"
#include <stdlib.h>

#define PRNAM "fboctree"
#define BIGVAL 0x10000000

typedef unsigned char UBYTE;
typedef unsigned long ULONG;

typedef struct node *OCTREE;

struct node
{
	UBYTE leaf;
	UBYTE level;
	UBYTE color;
	UBYTE rgb[3];
	UBYTE sub_count;
	ULONG pixels_low;
	ULONG pixels_high;
	ULONG red_sum_low;
	ULONG red_sum_high;
	ULONG green_sum_low;
	ULONG green_sum_high;
	ULONG blue_sum_low;
	ULONG blue_sum_high;
	OCTREE next_reduceable;
	OCTREE sub[8];
	OCTREE nlist;
	int index;
};

void search_colors(OCTREE tree);
void scan_large_tree(OCTREE tree);
void flip_image(FBM *image);
void clr_quantize(FBM *input, FBM *output, int colors);
void add_to_octree(FBM *image);
void add_to_large_octree(FBM *image);
void reduce_large_octree();

int write_bitmap (FBM *image,FILE *wfile, int type);
int read_bitmap (FBM *image, char *fname);
int alloc_fbm (FBM *image);
void search_tree(OCTREE tree);
int out_tree(char *treefile);
int input_tree(char *treefile, int add_flag);
int check_kennung(FILE *fp);
void print_usage();

#define MAXDEPTH 7

#define RED 0
#define GREEN 1
#define BLUE 2

#define MAXCOLORS 256
#define MIN_NODE_LIMIT 16
#define MAX_NODE_LIMIT 1024

#define TRUE		1
#define FALSE		0

static UBYTE palette[MAXCOLORS][3];

static int octree_size;
static int reduce_level;
static int leaf_level;
static int max_colors;
static int node_limit;
static int large_octree_flag;
static int reduce_start;

static OCTREE basetree;
static OCTREE reduce_list[MAXDEPTH + 1];
static int node_count[MAXDEPTH + 2];
static int reduce_count[MAXDEPTH + 1];
static int tree_count;
static OCTREE tlist;
static UBYTE b_field[]={128,64,32,16,8,4,2,1};

static double color_factor;
static int color_count;

#define MAX_KENN 20
static char Kennung[]="#OCTREE_v0.2a";
static char LargeKennung[]="#LARGE_OCTREE_v3";

#define SCANINT(x,y) \
if (sscanf(x,"%d",y) != 1) \
{ fprintf(stderr,"Invalid number '%s' in argument %d\n",x,i);\
 exit(1); }

#define GETPARAM(x) \
pending_param = &(x); pp_flag = 1; ppb = ppa;

/****************************************************************
 * main
 ****************************************************************/

void main (int argc, char *argv[])
{ FBM input, output;	/* Images */
  int color_depth, reduce_dynamics;
  int outtype;
  int add_flag, single_flag, flip_flag, info_flag;
  int i, pp_flag;
  int *pending_param;
  char *tfname, *ppa, *ppb;

  /* fprintf(stderr,"This is a test version (v5)!\n"); */
  /* Clear pointers */

  input.bm = input.cm = (UBYTE *) NULL;
  output.bm = output.cm = (UBYTE *) NULL;

  /* Defaults */
  color_depth=8;
  max_colors=MAXCOLORS;
  node_limit=2 * MAXCOLORS;
  outtype = DEF_8BIT;
  add_flag=0;
  single_flag=0;
  tfname=NULL;
  large_octree_flag=1;
  flip_flag = 0;
  info_flag = 0;
  reduce_dynamics = MAXDEPTH;
  pp_flag = 0;
  pending_param = NULL;
  ppa = ppb = NULL;

  /* scan arguments */

  if (argc == 1)
    { print_usage(); exit(1); }

  for (i=1; i < argc; i++)
    {
      ppa=argv[i];
      if (pp_flag == 1)
	{
	  SCANINT(ppa,pending_param);
	  pp_flag = 0;
	}
      else if (*ppa == '-')
	{
	  while (*(++ppa) != '\0')
	    {
	      if (pp_flag == 1)
		{
		  SCANINT(ppa,pending_param);
		  pp_flag = 0;
		  break;
		}
	      switch (*ppa)
		{
		case 'c':   GETPARAM(max_colors); break;
		case 'd':   GETPARAM(color_depth); break;
		case 'n':   GETPARAM(node_limit); break;
		case 'r':   GETPARAM(reduce_dynamics); break;
		case 'A':   outtype = FMT_ATK; break;
		case 'B':   outtype = FMT_FACE; break;
		case 'F':   outtype = FMT_FBM; break;
		case 'G':   outtype = FMT_GIF; break;
		case 'I':   outtype = FMT_IFF; break;
		case 'L':   outtype = FMT_LEAF; break;
		case 'M':   outtype = FMT_MCP; break;
		case 'P':   outtype = FMT_PBM; break;
		case 'R':   outtype = FMT_RLE; break;
		case 'S':   outtype = FMT_SUN; break;
		case 'T':   outtype = FMT_TIFF; break;
		case 'X':   outtype = FMT_X11; break;
		case 'Z':   outtype = FMT_PCX; break;
		case 'a':   add_flag = 1; break;
		case 'f':   flip_flag = 1; break;
		case 'v':   info_flag = 1; break;
		case 'u':   large_octree_flag = 0; break;
		case 's':   single_flag = 1; break;
		case 'h':   print_usage(); exit(1);
		default:
		  fprintf(stderr,"Illegal option '%c' in argument %d\n",
			  *ppa,i);
		  exit(1);
		}
	    }
	}
      else if (tfname == NULL)
	{
	  tfname=ppa;
	}
      else
	{
	  fprintf(stderr,"Extra file name found in argument %d\n",i);
	  exit(1);
	}
    }

  if (pp_flag == 1)
    {
      fprintf(stderr,"Missing number at option '%c' in argument %d\n",
	      *ppb,(argc-1));
      exit(1);
    }

  /* Check options and parameters */

  if (tfname == NULL && single_flag == 0)
    {
      fprintf(stderr,"No octree file specified\n");
      exit (1);

    }

  if (add_flag == 1 && single_flag == 1)
    {
      fprintf(stderr,"Options 'a' and 's' cannot be combined\n");
      exit (1);
    }

  if (color_depth > 8 || color_depth < 2)
    { 
      fprintf (stderr, "%s can only handle 2..8 color depth\n",PRNAM);
      exit (1);
    }

  if (max_colors > 256 || max_colors < 9)
    {
      fprintf (stderr, "%s can only handle 9..256 colors\n",PRNAM);
      exit (1);
    }

  if (reduce_dynamics < 0) reduce_dynamics = 0;

  /* prepare empty octree */
  for (i=0; i <= MAXDEPTH; i++)
    {
      reduce_list[i]=NULL;
      node_count[i]=0;
      reduce_count[i]=0;
    }
  node_count[MAXDEPTH + 1]=0;

  basetree=NULL;
  reduce_level = MAXDEPTH;
  leaf_level = reduce_level + 1;

  octree_size=0;

  if (tfname != NULL)
    {
      if (input_tree(tfname, add_flag) != 0) exit(1);
      if ((large_octree_flag == 1) && (info_flag == 1))
	{
	  fprintf(stderr," Leaf level: %d\n Node count:", leaf_level);
	  for (i=0; i <= (MAXDEPTH + 1); i++)
	    {
	      fprintf(stderr," %d",node_count[i]);
	    }
	  fprintf(stderr,"\n");
	}
    }

  if ((node_limit < MIN_NODE_LIMIT) || (node_limit > MAX_NODE_LIMIT))
    {
      fprintf(stderr,"The node limit %d is out of range (%d ... %d)\n",
	      node_limit,MIN_NODE_LIMIT, MAX_NODE_LIMIT);
      exit (1);
    }

  if (! read_bitmap (&input, (char *) NULL))
  { exit (1); }

  if (input.hdr.planes != 3 || input.hdr.bits != 8)
  { fprintf (stderr, "fboctree can only handle 24bit RGB inputs\n");
    exit (1);
  }

  if (add_flag == 1 || single_flag == 1)
    {
      fprintf (stderr,"Scanning \"%s\" [%dx%d]\n",
	       input.hdr.title, input.hdr.cols, input.hdr.rows);
      if (large_octree_flag == 1)
	{
	  add_to_large_octree(&input);
	  if (info_flag == 1)
	    {
	      fprintf(stderr," Leaf level: %d\n Node count:", leaf_level);
	      for (i=0; i <= (MAXDEPTH + 1); i++)
		{
		  fprintf(stderr," %d",node_count[i]);
		}
	      fprintf(stderr,"\n");
	    }
	}
      else
	{
	  add_to_octree(&input);
	  if (info_flag == 1)
	    { fprintf(stderr," Leaf level: %d\n", leaf_level); }
	}
    }

  if (add_flag == 1)
    {
      tree_count = 0;
      tlist=NULL;
      search_tree(basetree);
      fprintf(stderr,"Total octree nodes : %d\n",tree_count);

      out_tree(tfname);
      exit(0);
    }

  fprintf (stderr,
	"Quantizing \"%s\" [%dx%d] with %d colors (%d bits/clr)\n",
	input.hdr.title, input.hdr.cols, input.hdr.rows, max_colors,
	color_depth);

  for (i = 0; i < MAXCOLORS; i++)          /* init palette */
    palette[i][0] = palette[i][1] = palette[i][2] = 0;

  /* color 0 is *no longer* always black !! */

  if (large_octree_flag == 1)
    { color_count = 0; }
  else
    { color_count = 1; }

  color_factor = (double) ((1 << color_depth) - 1) / 0xFF;

  if (large_octree_flag == 1)
    {
      reduce_start = leaf_level - 1;
      for (i = 1; i < leaf_level; i++)
	{
	  if (node_count[i] >= max_colors)
	    {
	      reduce_start = i - 1;
	      break;
	    }
	}

      reduce_level = reduce_start + reduce_dynamics;
      if (reduce_level >= leaf_level)
	{
	  reduce_level = leaf_level - 1;
	  reduce_start = reduce_level - reduce_dynamics;
	  if (reduce_start < 0) reduce_start = 0;
	}
      else
	{
	  leaf_level = reduce_level + 1;
	}

      if (info_flag == 1)
	{ fprintf(stderr,"Reduce from %d to %d\n",reduce_start,reduce_level);}

      scan_large_tree(basetree);
      while (octree_size > max_colors)
	{
	  /* fprintf(stderr,"%d\n",octree_size); */
	  reduce_large_octree();
	}
      if (info_flag == 1)
	{
	  fprintf(stderr,"Reduce count:");
	  for (i=0; i <= MAXDEPTH; i++)
	    {
	      fprintf(stderr," %d",reduce_count[i]);
	    }
	  fprintf(stderr,"\n");
	}
    }

  search_colors(basetree);
  if (info_flag == 1)
    fprintf(stderr,"Number of colors : %d\n",color_count);

  /* Now build header for output bit map */
  output.hdr = input.hdr;
  output.hdr.planes = 1;
  output.hdr.clrlen = 3 * MAXCOLORS;
  
  /* Allocate space for output */
  alloc_fbm (&output);

  clr_quantize (&input, &output, MAXCOLORS);

  if (flip_flag == 1) flip_image(&output);

  /* Write out the result */
  if (write_bitmap (&output, stdout, outtype))
  { exit (0); }

  exit (1);
}

/************************************************************************
 * print_usage                                                          *
 ************************************************************************/

void print_usage()
{
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,"  %s -s [<options>] < 24bit > mapped\n",PRNAM);
  fprintf(stderr,"  %s -a [<options>] octree < 24bit\n",PRNAM);
  fprintf(stderr,"  %s [<options>] octree < 24bit > mapped\n",PRNAM);

  fprintf(stderr,"Parameters:\n\
  octree        file containing the octree information\n\
  24bit         true color input image\n\
  mapped        mapped 8bit output image\n");

  fprintf(stderr,"Mode control options:\n\
  -s            quantize single image independently (no octree file)\n\
  -a            scan image and add info to octree file (no output image)\n\
  [default]     quantize image using previously generated octree file\n");

  fprintf(stderr,"Options:\n\
  -c<numcolors> set maximum number of colors in output image (def: 256)\n\
  -d<bits>      set color depth in output image (def: 8)\n\
  -f            flip output image vertically\n\
  -n<nodelimit> set maximum number of nodes for each tree level (def: 512)\n\
  -r<redlevels> set number of levels used to reduce octree (def: 8)\n\
  -u            use original octree algorithm instead of modified one\n\
  -v            write some statistical numbers to stderr\n\
  -<type>       set output format, type is one of [FGIS]\n");

}

/************************************************************************
 * search_colors                                                        *
 ************************************************************************/

void search_colors(OCTREE tree)
{
  int j;
  double dhelp0, dhelp1;

  if (tree == NULL) return;

  dhelp0=(double) tree->pixels_high * (double) BIGVAL
    +(double) tree->pixels_low;

  dhelp0=color_factor / dhelp0;

  dhelp1=
    (double) tree->red_sum_high * (double) BIGVAL
      +(double) tree->red_sum_low;
  tree->rgb[RED] = (char) (dhelp0 * dhelp1 + 0.5);

  dhelp1=
    (double) tree->green_sum_high * (double) BIGVAL
      +(double) tree->green_sum_low;
  tree->rgb[GREEN] = (char) (dhelp0 * dhelp1 + 0.5);

  dhelp1=
    (double) tree->blue_sum_high * (double) BIGVAL
      +(double) tree->blue_sum_low;
  tree->rgb[BLUE] = (char) (dhelp0 * dhelp1 + 0.5);

  if (tree->leaf || tree->level == leaf_level)
    {
      palette[color_count][RED] = tree->rgb[RED];
      palette[color_count][GREEN] = tree->rgb[GREEN];
      palette[color_count][BLUE] = tree->rgb[BLUE];

      tree->color = color_count;
      tree->leaf = TRUE;
      color_count++;
    }
  else
    {
      for (j = 0; j < 8; j++)
	search_colors(tree->sub[j]);
    }
}

/************************************************************************
 * scan_large_tree                                                       *
 ************************************************************************/

void scan_large_tree(OCTREE tree)
{
  int j;

  if (tree == NULL) return;
  if ((tree->level <= reduce_level) && 
      (tree->level >= reduce_start))
    {
      if (tree->sub_count > 0)
	{
	  tree->next_reduceable = reduce_list[0];
	  reduce_list[0] = tree;
	}
    }

  if (tree->level == leaf_level)
    {
      tree->leaf = TRUE;
      octree_size++;
    }
  else
    {
      tree->leaf = FALSE;
      for (j = 0; j < 8; j++)
	scan_large_tree(tree->sub[j]);
    }
}

/************************************************************************
 * add_to_large_octree
 ************************************************************************/

void add_to_large_octree(FBM *image)
{
  UBYTE b_red, b_green, b_blue, branch, btest;
  UBYTE *p_red, *p_green, *p_blue;
  OCTREE tree, *p_tree;
  int i, depth, new_flag;

  p_red = image->bm;
  p_green = &p_red[image->hdr.plnlen];
  p_blue = &p_green[image->hdr.plnlen];

  for (i=0; i < image->hdr.plnlen; i++)
    {
      b_red = *(p_red++);
      b_green = *(p_green++);
      b_blue = *(p_blue++);

      p_tree = &basetree;
      new_flag = 0;

      for (depth = 0; depth <= leaf_level; depth++)
	{
	  if (*p_tree == NULL)            /* init new node */
	    {
	      tree = (OCTREE) calloc(1, sizeof(struct node));
	      if (tree == NULL)
		{
		  printf("Out of memory");
		  exit(1);
		}
	      tree->level = depth;
	      (node_count[depth])++;
	      new_flag = 1;
	      *p_tree = tree;
	    }
	  else
	    tree = *p_tree;

	  tree->pixels_low++;
	  tree->red_sum_low += b_red;
	  tree->green_sum_low += b_green;
	  tree->blue_sum_low += b_blue;

	  if (depth < leaf_level)
	    {
	      btest=b_field[depth];
	      branch=((b_red & btest) == 0) ? (UBYTE) 0 : (UBYTE) 4;
	      if ((b_green & btest) != 0) branch += (UBYTE) 2;
	      if ((b_blue & btest) != 0) branch += (UBYTE) 1;

	      if (tree->sub[branch] == NULL)
		tree->sub_count++;
	      p_tree=&(tree->sub[branch]);
	    }
	}

      if (new_flag)
	{
	  for (depth = 0; depth < leaf_level; depth++)
	    {
	      if (node_count[depth] >= node_limit)     /* reduce octree */
		{
		  leaf_level=depth;
		  break;
		}
	    }
	}
    }

  for (depth = (leaf_level+1); depth <= (MAXDEPTH+1) ;depth++)
    node_count[depth] = 0;
}

/************************************************************************
 * reduce_large_octree
 ************************************************************************/

void reduce_large_octree()
{
  int i, flag, depth;
  ULONG min_high, min_low;
  OCTREE tree, tree_before, next_tree, sub_tree;

  tree = reduce_list[0];
  if (tree == NULL)
    {
      fprintf(stderr,"Internal error: ZERO reduce\n");
      exit(1);
    }

  tree_before=NULL;

  min_high=tree->pixels_high;
  min_low=tree->pixels_low;

  while (1)
    {
      next_tree=tree->next_reduceable;
      if (next_tree == 0) break;

      flag = 1;
      for (i=0; i < 8; i++)
	{
	  sub_tree=next_tree->sub[i];
	  if (sub_tree != NULL)
	    { if (sub_tree->leaf == 0) {flag = 0;} }
	}

      if (flag == 1)
	{
	  if ((next_tree->pixels_high < min_high) ||
	      ((next_tree->pixels_high == min_high) &&
	       (next_tree->pixels_low < min_low)))
	    {
	      tree_before=tree;
	      min_high=next_tree->pixels_high;
	      min_low=next_tree->pixels_low;
	    }
	}
      tree=next_tree;
    }

  if (tree_before == NULL)
    {
      tree = reduce_list[0];
      reduce_list[0] = reduce_list[0]->next_reduceable;
    }
  else
    {
      tree = tree_before->next_reduceable;
      tree_before->next_reduceable = tree->next_reduceable;
    }

  tree->leaf = 1;
  octree_size = octree_size - tree->sub_count + 1;
  depth = tree->level;
  (reduce_count[depth])++;
}

/************************************************************************
 * add_to_octree                                                        *
 ************************************************************************/

void add_to_octree(FBM *image)
{
  UBYTE b_red, b_green, b_blue, branch, btest;
  UBYTE *p_red, *p_green, *p_blue;
  OCTREE tree, *p_tree;
  int i, depth;

  p_red = image->bm;
  p_green = &p_red[image->hdr.plnlen];
  p_blue = &p_green[image->hdr.plnlen];

  for (i=0; i < image->hdr.plnlen; i++)
    {
      b_red = *(p_red++);
      b_green = *(p_green++);
      b_blue = *(p_blue++);

      p_tree = &basetree;
      depth = 0;

      while (p_tree != NULL)
	{
	  if (*p_tree == NULL)            /* init new node */
	    {
	      tree = (OCTREE) calloc(1, sizeof(struct node));
	      if (tree == NULL)
		{
		  printf("Out of memory");
		  exit(1);
		}
	      tree->level = depth;
	      tree->leaf = (depth >= leaf_level);
	      if (tree->leaf) octree_size++;
	      *p_tree = tree;
	    }
	  else
	    tree = *p_tree;

	  tree->pixels_low++;
	  tree->red_sum_low += b_red;
	  tree->green_sum_low += b_green;
	  tree->blue_sum_low += b_blue;

	  if (tree->leaf == FALSE && depth < leaf_level)
	    {
	      btest=b_field[depth];
	      branch=((b_red & btest) == 0) ? (UBYTE) 0 : (UBYTE) 4;
	      if ((b_green & btest) != 0) branch += (UBYTE) 2;
	      if ((b_blue & btest) != 0) branch += (UBYTE) 1;

	      if (tree->sub[branch] == NULL)
		{
		  tree->sub_count++;
		  if (tree->sub_count == 2)   /* add to reduce_list */
		    {
		      tree->next_reduceable = reduce_list[depth];
		      reduce_list[depth] = tree;
		    }
		}
	      p_tree=&(tree->sub[branch]);
	      depth++;
	    }
	  else
	    p_tree = NULL;
	}

      if (octree_size >= max_colors)
	{
	  depth = reduce_level;
	  while (reduce_list[depth] == NULL) depth--;
	  /* reduce_octree(depth); */
	  tree = reduce_list[depth];
	  reduce_list[depth] = reduce_list[depth]->next_reduceable;
	  tree->leaf = 1;
	  octree_size = octree_size - tree->sub_count + 1;
	  if (depth < reduce_level)
	    {
	      reduce_level = depth;
	      leaf_level = reduce_level + 1;
	    }
	}
    }
}


/************************************************************************
 * flip_image
 ************************************************************************/

void flip_image(FBM *image)
{
  UBYTE *p1, *p2, *swap;
  int i, rows, rowlen;

  rows = image->hdr.rows;
  rowlen = image->hdr.rowlen;

  swap = (UBYTE *) malloc(rowlen);

  p1 = image->bm;
  p2 = p1 + ((rows - 1) * rowlen);

  for (i = 1; i < rows; i += 2)
    {
      memcpy(swap, p1, rowlen);
      memcpy(p1, p2, rowlen);
      memcpy(p2, swap, rowlen);
      p1 += rowlen;
      p2 -= rowlen;
    }
}

/************************************************************************
 * clr_quantize                                                         *
 ************************************************************************/

void clr_quantize(FBM *input, FBM *output, int colors)
{
  UBYTE b_red, b_green, b_blue, branch, btest;
  UBYTE *p_red, *p_green, *p_blue, *p_output;
  OCTREE tree, next_tree, sub_tree;
  int i,j,jmin,d1,d2,d3,dtest,dmin,unexpected;

  p_red = input->bm;
  p_green = &p_red[input->hdr.plnlen];
  p_blue = &p_green[input->hdr.plnlen];
  p_output = output->bm;

  unexpected=0;

  for (i=0; i < input->hdr.plnlen; i++)
    {
      b_red = *(p_red++);
      b_green = *(p_green++);
      b_blue = *(p_blue++);

      tree=basetree;
      next_tree=NULL;

      while (tree->leaf == 0)
	{
	  btest=b_field[tree->level];
	  branch=((b_red & btest) == 0) ? (UBYTE) 0 : (UBYTE) 4;
	  if ((b_green & btest) != 0) branch += (UBYTE) 2;
	  if ((b_blue & btest) != 0) branch += (UBYTE) 1;
	  next_tree=tree->sub[branch];
	  if (next_tree == NULL) break;
	  tree=next_tree;
	}

      if (next_tree == 0)
	{
	  unexpected++;
	  while (tree->leaf == 0)
	    {
	      jmin=-1;
	      dmin=0;
	      for (j=0; j < 8; j++)
		{
		  sub_tree=tree->sub[j];
		  if (sub_tree == NULL) continue;

		  d1=abs((int)sub_tree->rgb[RED] - (int)b_red);
		  d2=abs((int)sub_tree->rgb[GREEN] - (int)b_green);
		  d3=abs((int)sub_tree->rgb[BLUE] - (int)b_blue);

		  dtest=(d1 > d2) ? d1 : d2;
		  dtest=(d3 > dtest) ? d3 : dtest;

		  if ((jmin == -1) || (dtest < dmin))
		    {
		      dmin=dtest;
		      jmin=j;
		    }
		}
	      if (jmin == -1)
		{
		  fprintf(stderr,"W: %d\n",i);
		  break;
		}
	      tree=tree->sub[jmin];
	    }
	}

      *(p_output++) = tree->color;
    }

  p_red = output->cm;
  p_green = &p_red[colors];
  p_blue = &p_green[colors];

  for (i=0; i < colors; i++)
    {
      *(p_red++) = palette[i][RED];
      *(p_green++) = palette[i][GREEN];
      *(p_blue++) = palette[i][BLUE];
    }

  if (unexpected > 0)
    fprintf(stderr,"INFO: %d unexpected pixel(s)\n",unexpected);

}


/************************************************************************
 * search tree                                                          *
 ************************************************************************/

void search_tree(OCTREE tree)
{
  OCTREE subtree;
  int i;

  tree->index=tree_count++;
  tree->nlist=tlist;
  tlist=tree;

  while (tree->pixels_low >= BIGVAL)
    {
      tree->pixels_low -= BIGVAL;
      tree->pixels_high++;
    }

  while (tree->red_sum_low >= BIGVAL)
    {
      tree->red_sum_low -= BIGVAL;
      tree->red_sum_high++;
    }

  while (tree->green_sum_low >= BIGVAL)
    {
      tree->green_sum_low -= BIGVAL;
      tree->green_sum_high++;
    }

  while (tree->blue_sum_low >= BIGVAL)
    {
      tree->blue_sum_low -= BIGVAL;
      tree->blue_sum_high++;
    }

  if (tree->leaf == FALSE && tree->level < leaf_level)
    {
      for (i=0; i < 8; i++)
	{
	  subtree=tree->sub[i];
	  if (subtree != NULL)
	    search_tree(subtree);
	}
    }
}


/************************************************************************
 * out_tree                                                             *
 ************************************************************************/

int out_tree(char *tfname)
{
  FILE *fopen(), *output;
  OCTREE twork, thelp;
  int iwas,isub,i;

  fprintf(stderr,"Write octree file `%s'\n",tfname);

  if ((output = fopen(tfname, "w")) == NULL)
    {
      fprintf(stderr,"Error opening octree-file `%s'\n",tfname);
      return(1);
    }

  if (large_octree_flag == 1)
    { 
      fprintf(output,"%s\n",LargeKennung);
      fprintf(output,"%d %d\n",tree_count,node_limit);
      fprintf(output,"%d\n",leaf_level);
    }
  else
    {
      fprintf(output,"%s\n",Kennung);
      fprintf(output,"%d %d\n",tree_count,max_colors);
      fprintf(output,"%d %d %d\n",octree_size,reduce_level,leaf_level);
    }

  if (large_octree_flag != 1)
    {
      for (i=0; i <= MAXDEPTH; i++)
	{
	  thelp=reduce_list[i];
	  iwas=(thelp == NULL) ? 0 : (thelp->index+1);
	  fprintf(output," %X",iwas);
	}
    }
  else
    {
      for (i=0; i <= (MAXDEPTH + 1); i++)
	{
	  iwas=node_count[i];
	  fprintf(output," %X",iwas);
	}
    }

  fprintf(output,"\n");

  twork=tlist;
  while (twork != NULL)
    {
      if (twork->leaf == FALSE)
	isub=twork->sub_count;
      else
	isub=0;

      if (large_octree_flag != 1)
	{
	  fprintf(output,"%d %d %d",
		  (int) twork->leaf,
		  (int) twork->level,
		  isub);
	}
      else
	{
	  fprintf(output,"%d %d",
		  (int) twork->level,
		  isub);
	}

      fprintf(output,
	      " %lX %lX %lX %lX %lX %lX %lX %lX",
	      twork->pixels_low,
	      twork->pixels_high,
	      twork->red_sum_low,
	      twork->red_sum_high,
	      twork->green_sum_low,
	      twork->green_sum_high,
	      twork->blue_sum_low,
	      twork->blue_sum_high);

      if (isub != 0)
	{
	  for (i=0; i< 8; i++)
	    {
	      thelp=twork->sub[i];
	      iwas=(thelp == NULL) ? 0 : (thelp->index+1);
	      fprintf(output," %X",iwas);
	    }
	}

      if (large_octree_flag != 1)
	{
	  thelp=twork->next_reduceable;
	  iwas=(thelp == NULL) ? 0 : (thelp->index+1);

	  fprintf(output," %X\n",iwas);
	}
      else
	{
	  fprintf(output,"\n");
	}
      twork=twork->nlist;
    }

  fclose(output);
  return(0);
}

/************************************************************************
 * input_tree                                                             *
 ************************************************************************/

int input_tree(char *tfname, int add_flag)
{
  FILE *fopen(), *input;
  OCTREE twork;
  int n,i,iwas,il;

  n=0;
  if ((input = fopen(tfname, "r")) == NULL)
    {
      if (add_flag == 1)
	{
	  fprintf(stderr,"New octree file `%s'\n",tfname);
	  return(0);
	}
      else
	{
	  fprintf(stderr,"Can't open octree file `%s'\n",tfname);
	  return(1);
	}
    }

  fprintf(stderr,"Read octree file `%s'\n",tfname);
  fflush(stderr);

  il=1;
  if (check_kennung(input) != 1) goto Rfail; il++;

  if (fscanf(input,"%d",&tree_count) != 1) goto Rfail; il++;

  if (large_octree_flag != 1)
    {
      if (fscanf(input,"%d",&max_colors) != 1) goto Rfail; il++;

      if (fscanf(input,"%d",&octree_size) != 1) goto Rfail; il++;
      if (fscanf(input,"%d",&reduce_level) != 1) goto Rfail; il++;
      if (fscanf(input,"%d",&leaf_level) != 1) goto Rfail; il++;
    }
  else
    {
      if (fscanf(input,"%d",&node_limit) != 1) goto Rfail; il++;
    }

  if (fscanf(input,"%d",&leaf_level) != 1) goto Rfail; il++;

  tlist=(OCTREE) calloc(tree_count, sizeof(struct node));
  if (tlist == NULL)
    {
      fprintf(stderr,"Out of memory\n");
      fclose(input);
      return(1);
    }

  if (large_octree_flag != 1)
    {
      for (i=0; i <= MAXDEPTH; i++)
	{
	  if (fscanf(input,"%X",&iwas) != 1) goto Rfail; il++;
	  if (iwas > tree_count)
	    {
	      fprintf(stderr,"Bad octree file (a): %d %d\n",i,iwas);
	      goto Rfail;
	    }
	  if (iwas > 0)
	    reduce_list[i]=&tlist[iwas-1];
	  else
	    reduce_list[i]=NULL;
	}
    }
  else
    {
      for (i=0; i <= (MAXDEPTH+1); i++)
	{
	  if (fscanf(input,"%X",&iwas) != 1) goto Rfail; il++;
	  node_count[i]=iwas;
	}
    }

  il=0;
  for (n=tree_count-1; n >= 0; n--)
    {
      twork=&tlist[n];
      twork->index = 0;
      if (large_octree_flag != 1)
	{ if (fscanf(input,"%d",&iwas) != 1) goto Rfail; twork->leaf = iwas; }

      if (fscanf(input,"%d",&iwas) != 1) goto Rfail; twork->level = iwas;
      if (fscanf(input,"%d",&iwas) != 1) goto Rfail; twork->sub_count = iwas;

      if (fscanf(input,"%lX %lX %lX %lX %lX %lX %lX %lX",
		 &(twork->pixels_low),
		 &(twork->pixels_high),
		 &(twork->red_sum_low),
		 &(twork->red_sum_high),
		 &(twork->green_sum_low),
		 &(twork->green_sum_high),
		 &(twork->blue_sum_low),
		 &(twork->blue_sum_high)) != 8) goto Rfail;
      
      if (twork->sub_count != 0)
	{
	  for (i=0; i< 8; i++)
	    {
	      if (fscanf(input,"%X",&iwas) != 1) goto Rfail;
	      if (iwas > tree_count)
		{
		  fprintf(stderr,"Bad octree file (b): %d %d\n",n,iwas);
		  goto Rfail;
		}
	      if (iwas > 0)
		twork->sub[i] = &tlist[iwas-1];
	      else
		twork->sub[i] = NULL;
	    }
	}
      else
	{
	  for (i=0; i< 8; i++)
	    twork->sub[i] = NULL;
	}

      if (large_octree_flag != 1)
	{
	  if (fscanf(input,"%X",&iwas) != 1) goto Rfail;
	  if (iwas > tree_count)
	    {
	      fprintf(stderr,"Bad octree file (c): %d %d\n",n,iwas);
	      goto Rfail;
	    }
	  if (iwas > 0)
	    twork->next_reduceable = &tlist[iwas-1];
	  else
	    twork->next_reduceable = NULL;
	}
    }

  fclose(input);

  basetree=tlist;
  return(0);

Rfail:

  if (il == 0)
    { fprintf(stderr,"Error at octree node %d in file `%s'\n",n,tfname); }
  else
    { fprintf(stderr,"Error in header of file `%s' at pos %d\n",tfname,il); }
  fclose(input);
  return(1);
}

/****************************************************************
 * check_kennung
 ****************************************************************/

int check_kennung(FILE *fp)
{
  int c, i, k;
  char KennTest[MAX_KENN];

  i=0;
  while (1)
    {
      c=getc(fp);
      if (i < MAX_KENN) KennTest[i]=c;
      if ((c == '\n') || (c == EOF))
	{
	  if (i < MAX_KENN)
	    KennTest[i]='\0';
	  else
	    KennTest[MAX_KENN-1]='\0';
	  break;
	}
      i++;
    }

  k=0;
  if (strcmp(Kennung,KennTest) == 0) { k=1;}
  else if (strcmp(LargeKennung,KennTest) == 0) {k=2;}

  if (k == 0)
    {
      fprintf(stderr,"Unrecognized octree file format\n");
      return(0);
    }
  else if ((k == 1) && (large_octree_flag == 1))
    {
      fprintf(stderr,"Octree file contains normal version of octree\n");
      fprintf(stderr,"This is only useful with '-o' option\n");
      return(0);
    }
  else if ((k == 2) && (large_octree_flag != 1))
    {
      fprintf(stderr,"Octree file contains large version of octree\n");
      fprintf(stderr,"This is incompatible with '-o' option\n");
      return(0);
    }

  return(1);
}
